On 7/11/20 3:26 am, Henrik Rydberg wrote:
>>> I can't guarantee it won't break older machines which is why I've asked for 
>>> help testing it. I only have a MacbookPro 11,1 and an iMac 12,2. It fixes 
>>> both of those.
>>>
>>> Help testing would be much appreciated.
>>
>> I see, this makes much more sense. I may be able to run some tests tonight. 
>> Meanwhile, looking at the patch, the status variable in send_command looks 
>> superfluous now that there is a wait_status() before it.
> 
> Ok, it took some time to get the machines up to speed, but so far I have 
> managed to run some tests on an MacBookAir1,1. I only managed to upgrade up 
> to 4.15, so I had to revert the inputdev polling patch, but the rest applied 
> without problems. I recovered an old test program I used to use (attached), 
> and checked for failures and reads per second
> 
> *** hwmon: (applesmc) switch to using input device polling mode
> 
> At this point in the tree, with this reverted, I see 0 failures and 55 reads 
> per second.
> 
> *** hwmon: (applesmc) avoid overlong udelay()
> 
> With this applied, I see 0 failures and 16 reads per second.
> 
> *** hwmon: (applesmc) check status earlier.
> 
> With this applied, I see 0 failures and 16 reads per second.
> 
> *** (HEAD -> stable) applesmc: Re-work SMC comms v2
> 
> With this applied, the kernel hangs in module probe, and the kernel log is 
> flooded with read failures.
> 
> So as it stands, it does not work at all. I will continue to check another 
> machine, and see if I can get something working.
> 
> Henrik


G'day Heinrik,

If you could try this earlier version which still had all the failure logging 
it in we might be able to get a handle on the failure.

Regards,
Brad

---
diff --git a/drivers/hwmon/applesmc.c b/drivers/hwmon/applesmc.c
index a18887990f4a..22cc5122ce9a 100644
--- a/drivers/hwmon/applesmc.c
+++ b/drivers/hwmon/applesmc.c
@@ -42,6 +42,11 @@
 
 #define APPLESMC_MAX_DATA_LENGTH 32
 
+/* Apple SMC status bits from VirtualSMC */
+#define SMC_STATUS_AWAITING_DATA  0x01  ///< Data waiting to be read
+#define SMC_STATUS_IB_CLOSED      0x02  /// A write is pending / will ignore 
input
+#define SMC_STATUS_BUSY           0x04  ///< Busy in the middle of a command.
+
 /* wait up to 128 ms for a status change. */
 #define APPLESMC_MIN_WAIT      0x0010
 #define APPLESMC_RETRY_WAIT    0x0100
@@ -151,65 +156,77 @@ static unsigned int key_at_index;
 static struct workqueue_struct *applesmc_led_wq;
 
 /*
- * wait_read - Wait for a byte to appear on SMC port. Callers must
- * hold applesmc_lock.
+ * Wait for specific status bits with a mask on the SMC
+ * Used before and after writes, and before reads
  */
-static int wait_read(void)
+
+static int wait_status(u8 val, u8 mask)
 {
        unsigned long end = jiffies + (APPLESMC_MAX_WAIT * HZ) / USEC_PER_SEC;
        u8 status;
        int us;
 
        for (us = APPLESMC_MIN_WAIT; us < APPLESMC_MAX_WAIT; us <<= 1) {
-               usleep_range(us, us * 16);
                status = inb(APPLESMC_CMD_PORT);
-               /* read: wait for smc to settle */
-               if (status & 0x01)
+               if ((status & mask) == val)
                        return 0;
                /* timeout: give up */
                if (time_after(jiffies, end))
                        break;
-       }
-
-       pr_warn("wait_read() fail: 0x%02x\n", status);
+               usleep_range(us, us * 16);
+               }
+       pr_warn("wait_status timeout: 0x%02x, 0x%02x, 0x%02x\n", status, val, 
mask);
        return -EIO;
 }
 
 /*
  * send_byte - Write to SMC port, retrying when necessary. Callers
  * must hold applesmc_lock.
+ * Parameter skip must be true on the last write of any
+ * command or it'll time out.
  */
-static int send_byte(u8 cmd, u16 port)
+
+static int send_byte_data(u8 cmd, u16 port, bool skip)
 {
-       u8 status;
-       int us;
-       unsigned long end = jiffies + (APPLESMC_MAX_WAIT * HZ) / USEC_PER_SEC;
+       u8 wstat = SMC_STATUS_BUSY;
 
+       if (skip)
+               wstat = 0;
+       if (wait_status(SMC_STATUS_BUSY,
+       SMC_STATUS_BUSY | SMC_STATUS_IB_CLOSED))
+               goto fail;
        outb(cmd, port);
-       for (us = APPLESMC_MIN_WAIT; us < APPLESMC_MAX_WAIT; us <<= 1) {
-               usleep_range(us, us * 16);
-               status = inb(APPLESMC_CMD_PORT);
-               /* write: wait for smc to settle */
-               if (status & 0x02)
-                       continue;
-               /* ready: cmd accepted, return */
-               if (status & 0x04)
-                       return 0;
-               /* timeout: give up */
-               if (time_after(jiffies, end))
-                       break;
-               /* busy: long wait and resend */
-               udelay(APPLESMC_RETRY_WAIT);
-               outb(cmd, port);
-       }
-
-       pr_warn("send_byte(0x%02x, 0x%04x) fail: 0x%02x\n", cmd, port, status);
+       if (!wait_status(wstat,
+       SMC_STATUS_BUSY))
+               return 0;
+fail:
+       pr_warn("send_byte_data(0x%02x, 0x%04x) fail\n", cmd, 
APPLESMC_CMD_PORT);
        return -EIO;
 }
 
+/*
+ * send_command - Write a command to the SMC. Callers must hold applesmc_lock.
+ * If SMC is in undefined state, any new command write resets the state 
machine.
+ */
+
 static int send_command(u8 cmd)
 {
-       return send_byte(cmd, APPLESMC_CMD_PORT);
+       u8 status;
+
+       if (wait_status(0,
+       SMC_STATUS_IB_CLOSED)) {
+               pr_warn("send_command SMC was busy\n");
+               goto fail; }
+
+       status = inb(APPLESMC_CMD_PORT);
+
+       outb(cmd, APPLESMC_CMD_PORT);
+       if (!wait_status(SMC_STATUS_BUSY,
+       SMC_STATUS_BUSY))
+               return 0;
+fail:
+       pr_warn("send_cmd(0x%02x, 0x%04x) fail\n", cmd, APPLESMC_CMD_PORT);
+       return -EIO;
 }
 
 static int send_argument(const char *key)
@@ -217,7 +234,8 @@ static int send_argument(const char *key)
        int i;
 
        for (i = 0; i < 4; i++)
-               if (send_byte(key[i], APPLESMC_DATA_PORT))
+       /* Parameter skip is false as we always send data after an argument */
+               if (send_byte_data(key[i], APPLESMC_DATA_PORT, false))
                        return -EIO;
        return 0;
 }
@@ -233,13 +251,15 @@ static int read_smc(u8 cmd, const char *key, u8 *buffer, 
u8 len)
        }
 
        /* This has no effect on newer (2012) SMCs */
-       if (send_byte(len, APPLESMC_DATA_PORT)) {
+       if (send_byte_data(len, APPLESMC_DATA_PORT, false)) {
                pr_warn("%.4s: read len fail\n", key);
                return -EIO;
        }
 
        for (i = 0; i < len; i++) {
-               if (wait_read()) {
+               if (wait_status(SMC_STATUS_AWAITING_DATA | SMC_STATUS_BUSY,
+               SMC_STATUS_AWAITING_DATA | SMC_STATUS_BUSY |
+               SMC_STATUS_IB_CLOSED)) {
                        pr_warn("%.4s: read data[%d] fail\n", key, i);
                        return -EIO;
                }
@@ -250,7 +270,7 @@ static int read_smc(u8 cmd, const char *key, u8 *buffer, u8 
len)
        for (i = 0; i < 16; i++) {
                udelay(APPLESMC_MIN_WAIT);
                status = inb(APPLESMC_CMD_PORT);
-               if (!(status & 0x01))
+               if (!(status & SMC_STATUS_AWAITING_DATA))
                        break;
                data = inb(APPLESMC_DATA_PORT);
        }
@@ -263,20 +283,21 @@ static int read_smc(u8 cmd, const char *key, u8 *buffer, 
u8 len)
 static int write_smc(u8 cmd, const char *key, const u8 *buffer, u8 len)
 {
        int i;
+       u8 end = len-1;
 
        if (send_command(cmd) || send_argument(key)) {
                pr_warn("%s: write arg fail\n", key);
                return -EIO;
        }
 
-       if (send_byte(len, APPLESMC_DATA_PORT)) {
+       if (send_byte_data(len, APPLESMC_DATA_PORT, false)) {
                pr_warn("%.4s: write len fail\n", key);
                return -EIO;
        }
 
        for (i = 0; i < len; i++) {
-               if (send_byte(buffer[i], APPLESMC_DATA_PORT)) {
-                       pr_warn("%s: write data fail\n", key);
+               if (send_byte_data(buffer[i], APPLESMC_DATA_PORT, (i == end))) {
+                       pr_warn("%s: write data fail at %i\n", key, i);
                        return -EIO;
                }
        }
@@ -151,65 +156,77 @@ static unsigned int key_at_index;
 
 /*
- * send_byte - Write to SMC port, retrying when necessary. Callers
+ * send_byte_data - Write to SMC data port. Callers
  * must hold applesmc_lock.

Reply via email to